# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

"""
Functions passed to jinja templates
"""

import logging
from dateutil.parser import parse as parse_dt

from stag.utils import SortDate as _SortDate

from slugify import slugify

log = logging.getLogger(__name__)


def _is_date(val):
    return isinstance(val, date)


def strftime(val, format):
    if not val:
        return ""

    try:
        if isinstance(val, (int, float, str)):
            dt = parse_dt(val)
        else:
            dt = val
        return dt.strftime(format)
    except ValueError:
        return val


def isoformat(val):
    if isinstance(val, str):
        dt = parse_dt(val)
    else:
        dt = val
    return dt.isoformat()


def rfc822format(config):
    def flt(val):
        if isinstance(val, str):
            dt = parse_dt(val)
        else:
            dt = val
        return dt.strftime(f"%a, %d %b %Y %H:%M:%S {config.timezone}")

    return flt


def sorted_pages(pages, key=None, reverse=False):
    if key == "date":
        return sorted(pages, key=_SortDate, reverse=reverse)
    return sorted(pages, reverse=reverse)


def pagetype(pages, *filetypes):
    for page in pages:
        if page.output and any(page.output.type == ft for ft in filetypes):
            yield page


def update_env(env, config):
    globals = {"sorted_pages": sorted_pages, "slugify": slugify}
    filters = {
        "slugify": slugify,
        "strftime": strftime,
        "isoformat": isoformat,
        "rfc822format": rfc822format(config),
        "pagetype": pagetype,
    }

    env.globals.update(globals)
    env.filters.update(filters)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import itertools
import logging
import tomli

log = logging.getLogger(__name__)


def read_file(path):
    frontmatter = None
    frontmatter_parsed = False

    metadata = []
    content = []

    step = None  # forward declaration

    def detect_frontmatter(line):
        # front matter must be a line consisting of at least 3 the same characters
        return line and len(line) > 2 and all(ch == line[0] for ch in line)

    def parse_content(line):
        nonlocal content
        content.append(line)

    def parse_frontmatter(line):
        nonlocal frontmatter, frontmatter_parsed, metadata, step
        if frontmatter is None:  # possible only on the first line
            if detect_frontmatter(line):
                frontmatter = line
            else:
                step = parse_content
                step(line)
        elif line == frontmatter:
            frontmatter_parsed = True
            step = parse_content
        else:
            metadata.append(line)

    step = parse_frontmatter

    with open(path) as fd:
        for line in fd:
            line = line.rstrip("\r\n")
            step(line)

    if frontmatter and not frontmatter_parsed:
        log.warning(f"{path}: front matter recognised, but not closed.")

    if frontmatter_parsed:
        md = tomli.loads("\n".join(metadata))
        cn = "\n".join(content)
    else:
        md = {}
        if frontmatter is not None:
            cn = "\n".join(itertools.chain([frontmatter], content))
        else:
            cn = "\n".join(content)

    return md, cn, frontmatter_parsed



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
import shutil


class CopyError(Exception):
    pass


def copy_page(page, site):
    url = page.relurl.strip("/")
    output_path = os.path.join(site.config.output, url)

    if os.path.exists(output_path):
        raise CopyError(f"File exists: {output_path}")

    target_dir = os.path.dirname(output_path)
    os.makedirs(target_dir, exist_ok=True)
    shutil.copy(page.source.path, target_dir)



import pytest

from stag.utils import chdir
from stag.plugins.macros import MacrosConfig, resolve_macros
from stag.config import TemplateTable

from testutils import add_md_file, contents


@pytest.fixture(autouse=True)
def macros_config(config):
    config.plugins.macros = MacrosConfig()
    config.plugins.macros.path = "macros"

    config.plugins.theme = TemplateTable()
    config.plugins.theme.name = "theme"
    return config


@pytest.fixture(autouse=True)
def default_macros(macros_config, tmp_path):
    macros_dir = tmp_path / macros_config.plugins.macros.path
    macros_dir.mkdir()

    macro_templ = macros_dir / "macros.html"
    macro_content = "{%macro bold(name)%}<b>{{name}}</b>{% endmacro %}"
    macro_templ.write_text(macro_content)


def test_macros(site, tmp_path):
    content = '{%from "macros.html" import bold%}**foo** {{ bold("bar") }}'
    render_exp = f"**foo** <b>bar</b>"

    mdfile = add_md_file(site, content)

    with chdir(tmp_path):
        resolve_macros(site)
        assert mdfile.input.content == render_exp



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
from functools import cache

from typing import Optional as _Optional

import attr
from jinja2 import Template, Environment, ChoiceLoader, FileSystemLoader, PackageLoader
from jinja2.exceptions import TemplateSyntaxError, TemplateNotFound

from stag.signals import signals
from stag.ecs import Page


class MacroError(Exception):
    pass


def raise_macro_error(msg):
    raise MacroError(msg)


def import_module(*a, **kw):
    import importlib

    return importlib.import_module(*a, **kw)


@attr.s(auto_attribs=True)
class MacrosConfig:
    path: _Optional[str] = None


def is_viable(page):
    return page.source and page.input and page.input.content


@cache
def get_env(macros_dir, theme_dir):
    builtin = os.path.join(os.path.dirname(os.path.abspath(__file__)), "macros")
    loaders = [FileSystemLoader(dir) for dir in [macros_dir, theme_dir] if dir]
    loaders.append(PackageLoader("stag.plugins", "macros"))
    env = Environment(
        loader=ChoiceLoader(loaders),
        trim_blocks=True,
        lstrip_blocks=True,
        extensions=["jinja2.ext.debug", "jinja2.ext.do"],
    )
    env.globals["raise"] = raise_macro_error
    env.globals["import_module"] = import_module

    return env


PAGES = set()


def input_created(page, component):
    if not is_viable(page):
        return

    global PAGES
    PAGES.add(page)


def mutate(page, site, env):
    try:
        template = env.from_string(page.input.content)
        page.input.content = template.render(page=page, site=site)
    except TemplateNotFound as e:
        raise MacroError(f"{page.source.relpath}: missing macro definitions file: {e}")
    except Exception as e:
        raise MacroError(f"{page.source.relpath}: {e}")


def resolve_macros(site):
    config = site.config
    myconfig = config.plugins.macros

    env = get_env(myconfig.path, config.template.name)
    env.globals["site"] = site
    env.globals["print"] = print

    for page in site.pages:
        if page.source and page.input:
            mutate(page, site, env)


def register_plugin():
    signals.site_init.connect(
        lambda s: s.config.update_plugin_table("macros", MacrosConfig()), weak=False
    )
    signals.readers_finished.connect(resolve_macros)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
import itertools

from typing import List

import attr
import tomli
import markdown

from stag.ecs import Page, Content, Metadata
from stag.signals import signals, condition
from stag.reading import readers
from stag.plugins._helpers import read_file


def is_md(page):
    return page.source and page.source.ext == "md"


def is_opened_md(page):
    return page.input and page.input.type == "md"


@attr.s(auto_attribs=True)
class MarkdownConfig:
    extensions: List[str] = attr.Factory(lambda: ["footnotes", "fenced_code", "smarty"])


def deduce_url(path):
    if path.filebase == "index":
        return path.reldirname
    return os.path.join(path.reldirname, path.filebase)


def read(page):
    if not is_md(page):
        return

    metadata, content, _ = read_file(page.source.path)
    page.metadata = Metadata(metadata)
    page.input = Content("md", content)


def generate(site):
    myconfig = site.config.plugins.markdown
    conv = markdown.Markdown(extensions=myconfig.extensions)

    for page in site.pages:
        if not is_opened_md(page):
            continue

        assert "title" in page.metadata, f"No title in {page.source.relpath}"
        html = conv.reset().convert(page.input.content)
        page.output = Content("html", html)


def register_plugin():
    signals.site_init.connect(
        lambda s: s.config.update_plugin_table("markdown", MarkdownConfig()), weak=False
    )
    signals.readers_init.connect(
        (lambda site: site.page_added.connect(read)), weak=False
    )
    signals.processors_init.connect(generate)
    readers.register_reader(deduce_url, lambda p: p.ext == "md")



import markdown

from stag.ecs import Page, Path, Content


def add_md_file(site, text, metadata=None, parse=False):
    if metadata is None:
        metadata = {}

    config = site.config
    html = markdown.markdown(text) if parse else None
    return site.make_page(
        "/page",
        source=Path("page/index.md", site.config.content),
        metadata=metadata,
        input=Content("md", text),
        output=Content("html", html),
    )


def contents(path):
    with open(path) as file:
        return file.read().strip()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
import sys
import argparse
import shutil
import logging
import pkgutil
from copy import deepcopy

from stag.config import read_config
from stag.ecs import Site, Path, Page
from stag.utils import chdir
from stag.signals import signals
from stag.reading import readers
from stag.writing import render
from stag._version import version

log = logging.getLogger(__name__)


def load_plugins(paths):
    for finder, name, ispkg in pkgutil.iter_modules(paths):
        if name.startswith("_"):
            continue

        found = finder.find_module(name)
        if found:
            mod = found.load_module()
            try:
                mod.register_plugin()
            except AttributeError as e:
                log.error(f"{e} - plugin will be disabled")


def build(args):
    config = read_config(args.config)
    override_config_with_commandline_args(config, args)

    log.info(f"Building site to {config.output}")

    shutil.rmtree(config.output, ignore_errors=True)
    os.makedirs(config.output)

    builtin = os.path.join(os.path.dirname(__file__), "plugins")
    load_plugins([builtin, config.pluginspath])
    signals.plugins_loaded.emit()

    site = Site(config=config)
    signals.site_init.emit(site)

    roots = [
        config.content,
        os.path.join(config.template.name, "static"),
        "static",
    ]

    signals.readers_init.emit(site)

    for root in roots:
        gather_files(root, site)

    signals.readers_finished.emit(site)

    signals.processors_init.emit(site)
    signals.processors_finished.emit(site)

    signals.rendering_init.emit(site)
    render(site)
    signals.rendering_finished.emit(site)

    signals.site_finished.emit(site)


def serve(args):
    from http.server import HTTPServer, SimpleHTTPRequestHandler

    config = read_config(args.config)
    override_config_with_commandline_args(config, args)

    build(args)
    with chdir(config.output):
        log.info(f"Running simple HTTP server on http://localhost:{args.port}.")
        server_address = ("", args.port)
        httpd = HTTPServer(server_address, SimpleHTTPRequestHandler)
        httpd.serve_forever()


def deduce_url(path):
    if path.filebase == "index":
        return path.reldirname
    else:
        slug, _ = os.path.splitext(path.basename)
        return os.path.join(path.reldirname, pat.filebase)


def gather_files(srcdir, site):
    for curdir, _, files in os.walk(srcdir):
        for f in files:
            path = Path(os.path.join(curdir, f), srcdir)
            relurl = readers.get_path(path)
            site.make_page(relurl, source=path)


def override_config_with_commandline_args(config, args):
    for name, val in args.__dict__.items():
        if val is not None and hasattr(config, name):
            setattr(config, name, args.__dict__[name])


def prepare_args(argv):
    parser = argparse.ArgumentParser(description="Simply Stupid Static Site Generator")
    parser.set_defaults(verbosity=logging.INFO)

    parser.add_argument(
        "-c",
        "--config",
        nargs="?",
        default="config.toml",
        help="path to stag's configuration file",
    )

    parser.add_argument(
        "-D",
        "--debug",
        action="store_const",
        const=logging.DEBUG,
        dest="verbosity",
        help="show debug messages",
    )

    parser.add_argument("--version", action="version", version=f"%(prog)s {version}")

    sp = parser.add_subparsers(required=True, dest="subcommands")

    sp_build = sp.add_parser("build")
    sp_build.add_argument("-o", "--output", help="output directory")
    sp_build.set_defaults(func=build)

    sp_serve = sp.add_parser("serve")
    sp_serve.add_argument("-o", "--output", help="output directory")
    sp_serve.add_argument("-p", "--port", type=int, default="8000", help="HTTP port")
    sp_serve.set_defaults(func=serve)

    return parser.parse_args(argv)


def main_(argv):
    args = prepare_args(argv)
    logging.basicConfig(format="%(message)s", level=args.verbosity)

    try:
        return args.func(args)
    except Exception as e:
        log.error(f"Critical error: {e}")
        if args.verbosity == logging.DEBUG:
            import pdb, traceback

            extype, value, tb = sys.exc_info()
            traceback.print_exc()
            pdb.post_mortem(tb)
            raise
        return 1


def main():
    return main_(sys.argv[1:])



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import attr

from typing import Optional, Callable, List

from stag.ecs import Path


@attr.s(auto_attribs=True)
class ContentReader:
    reader: Callable[[Path], str]
    condition: Optional[Callable[[Path], bool]] = None


@attr.s(auto_attribs=True)
class Readers:
    _readers: List[ContentReader] = attr.ib(factory=list)

    def register_reader(self, reader, condition=None):
        self._readers.append(ContentReader(reader, condition))

    def get_path(self, path):
        reader = next(
            (r.reader for r in self._readers if r.condition(path)), self._default_path
        )
        return reader(path)

    def _default_path(self, path):
        return path.relpath


readers = Readers()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
from collections import UserDict, UserList
from contextlib import contextmanager as _contextmanager
from datetime import date as _date_t
from datetime import datetime as _datetime_t


@_contextmanager
def chdir(path):
    cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


class SortDate:
    def __init__(self, page):
        self.ts = None
        if page.metadata:
            self.ts = page.metadata.get("timestamp")

    def __lt__(self, other):
        if self.ts and other.ts:
            return self.ts < other.ts
        return self.ts and not other.ts


class AttrDict(UserDict):
    def __getitem__(self, key):
        return attr_access(self.data[key])

    def __getattr__(self, key):
        try:
            return attr_access(getattr(self.data, key))
        except AttributeError:
            try:
                return self[key]
            except KeyError:
                raise AttributeError(key)


class AttrList(UserList):
    def __getitem__(self, key):
        return attr_access(self.data[key])


def attr_access(val):
    if isinstance(val, dict):
        return AttrDict(val)
    if isinstance(val, (tuple, list)):
        return AttrList(val)
    return val



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import logging

from typing import Optional, Callable, List

from stag.ecs import Path
from stag.writers.jinja import render_page, get_env
from stag.writers.copy import copy_page


log = logging.getLogger(__name__)


def is_static(page):
    return page.source and not page.input


def outputable(page):
    return page.output and page.metadata


def render(site):
    env = get_env(site.config.template.name, site.config)
    for page in site.pages:
        if outputable(page):
            render_page(page, site, env)
        elif is_static(page):
            copy_page(page, site)
        else:
            log.error(f"Couldn't deduce how to render page: {page.relurl}")



from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as f_:
    long_description = f_.read()


def main():
    setup(
        name="stag-ssg",
        description="Deadly simple static site generator",
        long_description=long_description,
        long_description_content_type="text/markdown",
        use_scm_version={"write_to": "src/stag/_version.py"},
        license="GPLv3+",
        author="Michał Góral",
        author_email="dev@goral.net.pl",
        url="https://git.goral.net.pl/mgoral/stag",
        platforms=["linux"],
        python_requires=">=3.7,<3.10",
        setup_requires=["setuptools_scm"],
        install_requires=[
            "attrs >=21, <22",
            "tomli >=1.2, <2.0",
            "jinja2 >=3 ,<4",
            "markdown >=3, <4",
            "python-dateutil >=2.8.0, <3",
            "python-slugify >=5, <6",
        ],
        extras_require={},
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Intended Audience :: End Users/Desktop",
            "Intended Audience :: Developers",
            "Intended Audience :: Information Technology",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Natural Language :: English",
            "Operating System :: POSIX",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Internet :: WWW/HTTP :: Site Management",
            "Topic :: Utilities",
        ],
        packages=find_packages("src"),
        package_dir={"": "src"},
        package_data={"stag": ["plugins/*.py", "plugins/macros/*.html"]},
        entry_points={
            "console_scripts": [
                "stag=stag.stag:main",
            ],
        },
    )


if __name__ == "__main__":
    main()



import os
import pytest

from stag.utils import chdir
from stag.config import TemplateTable
from stag.writers.jinja import render_page, get_env

from testutils import add_md_file, contents


@pytest.fixture(autouse=True)
def jinja_config(config):
    config.template = TemplateTable()
    config.template.name = "theme"
    return config


@pytest.fixture(autouse=True)
def default_templates(jinja_config, tmp_path):
    templates_dir = tmp_path / jinja_config.template.name
    templates_dir.mkdir()

    html_templ = templates_dir / "page.html"
    html_content = "<html><body>{{ content }}</body></html>"
    html_templ.write_text(html_content)
    print(html_templ)


def test_basic_gen(site, tmp_path):
    mdfile = add_md_file(site, "Content", parse=True)
    render_exp = f"<html><body>{mdfile.output.content}</body></html>"

    with chdir(tmp_path):
        env = get_env(site.config.template.name, site.config)
        render_page(mdfile, site, env)
        out = os.path.join(
            tmp_path, site.config.output, mdfile.relurl.strip("/"), "index.html"
        )
        assert contents(out) == render_exp
